package org.eclipse.fdt.internal.core.f95parser;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.Reader;

/**
 * FixedFormLexerPrepass preprocesses the input stream. It discards all
 * whitespaces and comment lines and concatenates continuation lines. Additional
 * it holds a mapping to the character-positions in the file (for correct
 * start/end line/col in the Token objects).
 * 
 * @author Dirk Rossow
 */
class FixedFormLexerPrepass {
	static final int inStart=0;
	static final int inHollerith=1;
	static final int inDblQuote=2;
	static final int inDblQuoteEnd=3;
	static final int inQuote=4;
	static final int inQuoteEnd=5;

	private int state = inStart;
    int hollerithLength = -2; //-1: hollerith could start, -2: hollerith cant start

	private int actLinePos = 0;
	private PreLexerLine actLine = null;
	private OffsetLineReader in;

	private DynamicIntArray lineMapping = new DynamicIntArray(1000);
	private DynamicIntArray columnMapping = new DynamicIntArray(1000);
	private DynamicIntArray offsetMapping = new DynamicIntArray(1000);
	
	private int EOFLinePos=0;
	private int EOFColPos=0;
	private int EOFOffsetPos=0;
	
	
	public FixedFormLexerPrepass(Reader in) {
		this.in = new OffsetLineReader(in);
	}
	
	public FixedFormLexerPrepass(InputStream in) {
		this(new InputStreamReader(in));
	}
	
	public int getLine(int absCharPos) {
		return lineMapping.get(absCharPos);
	}
	
	public int getColumn(int absCharPos) {
		return columnMapping.get(absCharPos);
	}
	
	public int getOffset(int absCharPos) {
		return offsetMapping.get(absCharPos);
	}

	public int read() throws Exception {
		int c = internalRead();
	//	System.out.print((char)c);
		return c;
	}

	private void markPosition(int line, int col, int offset) {
		lineMapping.pushBack(line);
		columnMapping.pushBack(col);
		offsetMapping.pushBack(offset);
	}
	
	private int internalRead() throws Exception {
		
		for (;;) {
			if (actLine==null) {
				actLine=getNextNonCommentLine();
				if (actLine==null) {
					markPosition(EOFLinePos,EOFColPos,EOFOffsetPos);
					return -1;
				}
				else actLinePos=0;
			} else if (actLinePos==actLine.length()) { //test if continuation-line follows, else send \n
				PreLexerLine prevLine=actLine;
				actLine=getNextNonCommentLine();
				if (actLine==null) { //end of file
					hollerithLength=-2;
					state=inStart;
					markPosition(prevLine.linePos,prevLine.length(),prevLine.offset+actLinePos+1);
					return '\n';
				} else if (actLine.type==PreLexerLine.CONTINUATION) {
					actLinePos=6;
				} else {
					actLinePos=0;
					hollerithLength=-2;
					state=inStart;
					markPosition(prevLine.linePos,prevLine.length(),prevLine.offset);
					return '\n';
				}
			}
			
			actLinePos = getNextSigPos(actLine,actLinePos);
			if (actLinePos<0) {
				actLinePos=actLine.length();
			} else {
				markPosition(actLine.linePos,actLinePos,actLine.offset+actLinePos);
				return actLine.charAt(actLinePos++);
			}
		}
	}

	private PreLexerLine getNextNonCommentLine() {
		for (;;) {
			PreLexerLine line = getNextLine();
			if (line==null) return null;
			if (line.type!=PreLexerLine.COMMENT) return line;
		}
	}

	private PreLexerLine getNextLine() {
		try {
			int actOffset=in.getOffset();
			String line = in.readLine();
			if (line==null) return null;
			EOFLinePos=in.getLineNumber()-1;
			EOFColPos=line.length();
			EOFOffsetPos=actOffset;
			PreLexerLine pll = new PreLexerLine(line,in.getLineNumber()-1,actOffset);
			return pll;
		} catch (IOException e) {
			return null;
		}
	}

	// return: -1 : end of line reached
	private int getNextSigPos(PreLexerLine line, int startPos) {

		  for (int charPos=startPos;charPos<line.length();++charPos)  {
			  char c = line.charAt(charPos);
			  
			if (state==inStart) {
				if (c==' ' || c=='\t') continue;
				else if (c=='!') {
					return -1; // rest of line is comment
				} else if (c=='\'') {
					hollerithLength=-1;
					state=inQuote;
				} else if (c=='\"') {
					hollerithLength=-1;
					state=inDblQuote;
				} else if ((c=='h') || (c=='H')) {
					if (hollerithLength>0) state=inHollerith;
					else if (hollerithLength<0) hollerithLength=-2;
				} else if (hollerithLength!=-2 && Character.isDigit(c)) {
					if (hollerithLength==-1) hollerithLength=Character.digit(c,10);
					else hollerithLength=hollerithLength*10+Character.digit(c,10);
				} else if (Character.isLetter(c) || c=='_') {
					hollerithLength=-2;
				} else {
					if (charPos==0) hollerithLength=-2;// ignore label at start of line
					else hollerithLength=-1;
				}
				return charPos;
				
			} else if (state==inQuote) {
				if (c=='\'') state=inQuoteEnd;
				return charPos;
			} else if (state==inQuoteEnd) {
				if (c=='\'') {
					state=inQuote;
					return charPos;
				} else {
					state=inStart;
					charPos--;
				}
			} else if (state==inDblQuote) {
				if (c=='\"') state=inQuoteEnd;
				return charPos;
			} else if (state==inDblQuoteEnd) {
				if (c=='\"') {
					state=inDblQuote;
					return charPos;
				} else {
					state=inStart;
					charPos--;
				}				
			} else if (state==inHollerith) {
				hollerithLength--;
				if (hollerithLength==0) state=inStart;
				return charPos;
			} else { //undefined state
				throw new RuntimeException("Undefined state in FixedFormPreLexer");
			}
		}
		return -1; //end of line reached
	}
}

class DynamicIntArray {
	final double RESIZEFAC=1.1;
	final int RESIZEADD=10;
	
	private int[] v=null;
	private int length=0;
	
	DynamicIntArray(int reserveSize) {
		ensureSize(reserveSize);
	}

	DynamicIntArray() {
	}

	int length() {
		return length;
	}
	
	int get(int pos) {
		if (pos<0 || pos>=length) {
			throw new ArrayIndexOutOfBoundsException(pos);
		}
		return v[pos];
	}
	
	void set(int pos, int value) {
		if (pos<0 || pos>=length) throw new ArrayIndexOutOfBoundsException(pos);
		v[pos]=value;
	}
	
	void pushBack(int value) {
		ensureSize(length+1);
		v[length]=value;
		length++;
	}
	
	private void ensureSize(int size) {
		if (v==null || v.length<size) {
			int[] newArray = new int[(int) (size*RESIZEFAC+RESIZEADD)];
			if (v!=null) System.arraycopy(v,0,newArray,0,length);
			v=newArray;
		}
	}
	
	
}

class PreLexerLine {
	static final int COLWIDTH=72;
	static final int COMMENT=0;
	static final int CONTINUATION=1;
	static final int STMT=2;
	
	final int linePos;
	final int offset;
	final int type;
	private final String lineText;
	
	
		
	PreLexerLine(String _lineText, int linePos, int offset) {
		/* truncate anything beyond 72 characters */	    
		if (_lineText.length()>COLWIDTH) {
			lineText=_lineText.substring(0,COLWIDTH);
		} else {
			lineText=_lineText;
		}
		this.linePos=linePos;
		this.offset=offset;
		 
		//determine type of line
		if (lineText.length()==0) this.type=COMMENT;
		else {
//check for f77-style comment
			char firstChar=lineText.charAt(0);			
			//ignore cpp-directives for now
			//ignore $ALIAS for now
			if (firstChar=='C' || firstChar=='c' 
				|| firstChar=='*' || firstChar=='!' 
					|| firstChar=='#' || firstChar=='$') type=COMMENT;

//check for continuation
			else {
				int idx=lineText.indexOf('\t');
				if (idx>=0 && idx<=5) { //stmt or comment
					String afterCol5 = lineText.substring(idx+1,lineText.length()).trim();
					if (afterCol5.length()<=0 || afterCol5.startsWith("!")) type=COMMENT;
					else type=STMT;
				} else {
					if (lineText.length()>=6 &&
						lineText.startsWith("     ") &&
						lineText.charAt(5)!=' ' && 
						lineText.charAt(5)!='0') type = CONTINUATION;
					else if (lineText.startsWith("      ")) {
						String trimmed = lineText.trim();
						if (trimmed.length()<=0 || trimmed.startsWith("!")) type=COMMENT;
						else type=STMT;
					} else type=STMT;
				}
			}
		}
	}
	 
	public int length() {
		return lineText.length();
	}
	
	public char charAt(int pos) {
		return lineText.charAt(pos);
	}
	

	public String toString() {
		return "Line "+linePos+": "+lineText; 
	}
}

class OffsetLineReader {
    LineNumberReader lnr;
    int offset=0;
	

	public OffsetLineReader(Reader arg0) {
		lnr = new LineNumberReader(arg0);
	}
	
	public String readLine() throws IOException {
		lnr.mark(100000);
		//find length of line (including terminals)
		int lineLength;
		
		// JO: lineLength should start at 0 if the line terminator is \n
		// or \r and 1 if it is \r\n
		lineLength = System.getProperty("line.separator").length() - 1;
		
		for(;;) {
			int c = lnr.read();
			lineLength++;
			if (c=='\n') {
				break;
			} else if (c=='\r') {
				int c2 = lnr.read();
				if (c2=='\n') {
					lineLength++;
				}
				break;
			} else if (c==-1) {
				lineLength--;
				break;
			}
		}
		lnr.reset();
		offset+=lineLength;
		return lnr.readLine();
	}
	
	public int getLineNumber() {
		return lnr.getLineNumber();
	}
	
	public int getOffset() {
		return offset;
	}
}
